
//---------------------------------"
// This file is generated by make."
// Any changes will be lost!"
//---------------------------------"

#define TOOLSET_SHORT	"@@TOOLSET-SHORT@@"
#define TOOLSET_LONG	"@@TOOLSET-LONG@@"
#define WIT_SHORT	"@@WIT-SHORT@@"
#define WIT_LONG	"@@WIT-LONG@@"
#define WWT_SHORT	"@@WWT-SHORT@@"
#define WWT_LONG	"@@WWT-LONG@@"
#define WDF_SHORT	"@@WDF-SHORT@@"
#define WDF_LONG	"@@WDF-LONG@@"
#define WFUSE_SHORT	"@@WFUSE-SHORT@@"
#define WFUSE_LONG	"@@WFUSE-LONG@@"
#define AUTHOR		"@@AUTHOR@@"

#define VERSION		"@@VERSION@@"
#define VERSION_NUM	"@@VERSION-NUM@@"
#define BETA_VERSION	@@BETA-VERSION@@

#define REVISION	"@@REV@@"
#define REVISION_NUM	@@REV-NUM@@
#define REVISION_NEXT	@@REV-NEXT@@

#define BINTIME		@@BINTIME@@
#define DATE		"@@DATE@@"
#define TIME		"@@TIME@@"
#define DATETIME	"@@DATE@@ @@TIME@@"

#define INSTALL_PATH	"@@INSTALL-PATH@@"
#define SHARE_PATH	"@@SHARE-PATH@@"
#define WIN_INSTALL_PATH "@@WIN-INSTALL-PATH@@"

#define URI_REPOS	"@@URI-REPOS@@/"
#define URI_VIEWVC	"@@URI-VIEWVC@@/"
#define URI_HOME	"@@URI-HOME@@"
#define URI_DOWNLOAD	"@@URI-DOWNLOAD@@"

