
//---------------------------------"
// This file is generated by make."
// Any changes will be lost!"
//---------------------------------"

#gdef wit$ = @map

#gdef wit$['TOOLSET_SHORT']		= "@@TOOLSET-SHORT@@"
#gdef wit$['TOOLSET_LONG']		= "@@TOOLSET-LONG@@"
#gdef wit$['WIT_SHORT']			= "@@WIT-SHORT@@"
#gdef wit$['WIT_LONG']			= "@@WIT-LONG@@"
#gdef wit$['WWT_SHORT']			= "@@WWT-SHORT@@"
#gdef wit$['WWT_LONG']			= "@@WWT-LONG@@"
#gdef wit$['WDF_SHORT']			= "@@WDF-SHORT@@"
#gdef wit$['WDF_LONG']			= "@@WDF-LONG@@"
#gdef wit$['WFUSE_SHORT']		= "@@WFUSE-SHORT@@"
#gdef wit$['WFUSE_LONG']		= "@@WFUSE-LONG@@"
#gdef wit$['AUTHOR']			= "@@AUTHOR@@"

#gdef wit$['VERSION']			= "@@VERSION@@"
#gdef wit$['VERSION_NUM']		= "@@VERSION-NUM@@"
#gdef wit$['BETA_VERSION']		= "@@BETA-VERSION@@"
#gdef wit$['BETA_SUFFIX']		= "@@BETA-SUFFIX@@"

#gdef wit$['REVISION']			= "@@REV@@"
#gdef wit$['REVISION_NUM']		= @@REV-NUM@@
#gdef wit$['REVISION_NEXT']		= @@REV-NEXT@@

#gdef wit$['BINTIME']			= @@BINTIME@@
#gdef wit$['DATE']			= "@@DATE@@"
#gdef wit$['TIME']			= "@@TIME@@"
#gdef wit$['DATETIME']			= "@@DATE@@ @@TIME@@"

