
/***************************************************************************
 *                         _______ _______ _______                         *
 *                        |  ___  |____   |  ___  |                        *
 *                        | |   |_|    / /| |   |_|                        *
 *                        | |_____    / / | |_____                         *
 *                        |_____  |  / /  |_____  |                        *
 *                         _    | | / /    _    | |                        *
 *                        | |___| |/ /____| |___| |                        *
 *                        |_______|_______|_______|                        *
 *                                                                         *
 *                            Wiimms SZS Tools                             *
 *                          https://szs.wiimm.de/                          *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This file is part of the SZS project.                                 *
 *   Visit https://szs.wiimm.de/ for project details and sources.          *
 *                                                                         *
 *   Copyright (c) 2011-2021 by Dirk Clemens <wiimm@wiimm.de>              *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   See file gpl-2.0.txt or http://www.gnu.org/licenses/gpl-2.0.txt       *
 *                                                                         *
 ***************************************************************************/

#ifndef USE_SHA_LIB
 #ifdef __CYGWIN__
    #define SHA1_ASM
    #ifdef WIIMM_INCLUDE_SSL_ASM
	#if defined(__x86_64__)
	    #include "sha1-cygwin64.s"
	#else
	    #include "sha1-586-cygwin.s"
	#endif
    #endif
 #elif defined(__linux__)
    #if defined(__i386__)
	#define SHA1_ASM
	#ifdef WIIMM_INCLUDE_SSL_ASM
	    #include "sha1-586-elf.s"
	#endif
    #elif defined(__x86_64__)
	#define SHA1_ASM
	#ifdef WIIMM_INCLUDE_SSL_ASM
	    #include "sha1-x86_64.s"
	#endif
    #endif
 #endif
#endif // !USE_SHA_LIB

